/*
 *	NET3:	Generic stream/socket interface for MAC protocols.
 *	
 *		Alan Cox, <A.Cox@swansea.ac.uk / gw4pts@gw4pts.ampr.org>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */


static void gs_callback1(struct sock *sk)
{
	if(!sk->dead)
		wake_up_interruptible(sk->sleep);
}

static void gs_callback2(struct sock *sk, int len)
{
	if(!sk->dead)
		wake_up_interruptible(sk->sleep);
}

static int gs_create(struct socket *sock, int protocol)
{
	struct sock *sk;
	sk=(struct sock *)kmalloc(sizeof(*sk), GFP_KERNEL);
	if(sk==NULL)
		return -ENOBUFS;
	sk->num=protocol;
	sk->reuse=0;
	if(!suser())
	{
		kfree_s((void *)sk, sizeof(*sk));
		return -EPERM;
	}
	
	switch(sock->type)
	{
		case SOCK_DGRAM:
		case SOCK_RAW:
			break;
		default:
			kfree_s((void *)sk,sizeof(*sk));
			return -ESOCKTNOSUPPORT;
	}
	sk->dead=0;
	sk->next=NULL;
	sk->broadcast=NULL;
	sk->rcvbuf=SK_RMEM_MAX;
	sk->sndbuf=SK_WMEM_MAX;
	sk->wmem_alloc=0;
	sk->rmem_alloc=0;
	sk->inuse=0;
	sk->shutdown=0;
	sk->prot=NULL;	/* So we use default free mechanisms */
	sk->err=0;
	skb_queue_head_init(&sk->receive_queue);
	skb_queue_head_init(&sk->write_queue);
	sk->send_head=NULL;
	skb_queue_head_init(&sk->back_log);
	sk->state=TCP_CLOSE;
	sk->socket=sock;
	sk->type=sock->type;
	sk->debug=0;
	sk->mtu=0;
	sk->state_change=gs_callback1;
	sk->data_ready=gs_callback2;
	sk->write_space=gs_callback1;
	sk->error_report=gs_callback1;
	sk->zapped=1;
	sk->opt.generic.family=AF_UNSPEC;
	sk->opt.generic.device[0]=0;
	return(0);
}

static int gs_dup(struct socket *newsock, struct socket *oldsock)
{
	return gs_create(newsock, oldsock->tyoe);
}

static int gs_release(struct socket *sock, struct socket *peer)
{
	struct sock *sk=(struct sock *)sock->data;
	if(sk==NULL)
		return;
	if(!sk->dead)
		sk->state_change(sk);
	sk->dead=1;
	sock->data=NULL;
	gs_destroy_socket(sk);
	return 0;
}

/*
 *	Generic mac protocol sockets are bound to physical device layers.
 */

static int gs_bind(struct socket *sock, struct sockaddr *umyaddr, int sockaddr_len)
{
	struct sock *sk=(struct sock *)sock->data;
	if(addr_len>=sizeof(short)+sizeof(sk->opt.generic.device))
		return -EINVAL;
	if(sk->opt.generic.bound)
		return -EINVAL;
	memcpy(sk->opt.generic.device,umyaddr.sa_data,sizeof(sk->opt.generic.device));
	sk->opt.generic.device[sizeof(sk->opt.generic.device)-1]=0;
	if(dev_find(sk->opt.generic.device)==NULL)
		return -ENODEV;
	sk->opt.generic.bound=1;
	return 0;
}
	
/*
 *	Connect is protocol specific - a totally generic layer just doesn't support it.
 */
 
static int gs_connect(struct socket *sock, struct sockaddr *umyaddr, int sockaddr_len, int flags)
{
	return -EOPNOTSUPP;
}


/*
 *	Generic sockets don't pair. Use streams loopback or something else pal.
 */
 
static int gs_socketpair(struct socket *sock1, struct socket *sock2)
{
	return -EOPNOTSUPP;
}

/*
 *	Datagram only - no accepts
 */

static int gs_accept(struct socket *sock, struct socket *newsock, int flags)
{
	return -EOPNOTSUPP;
}

/*
 *	Get the address name of our protocol layer.
 */
 
static int gs_getname(struct socket *sock, struct sockaddr *uaddr, int *usockaddr_len, int peer)
{	
	struct sock *sk=(struct sock *)sock->data;
	if(peer)
		return -ENOTCONN;
	uaddr->sa_family=AF_UNSPEC;	/* Hardware address */
	memcpy(uaddr.sa_data,sk->opt.generic.device, sizeof(uaddr.sa_data));
	*usockaddr_len=sizeof(struct sockaddr);
	return 0;
}

static int gs_select(struct socket *sock, int sel_type, select_table *wait)
{
	return datagram_select((struct sock *)sock->data, sel_type, wait);
}

static int gs_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
	int err;
	long amount=0;
	struct sock *sk=(struct sock *)sock->data;
	
	switch(cmd)
	{
		case TIOCOUTQ:
			err=verify_area(VERIFY_WRITE,(void *)arg,sizeof(unsigned long));
			if(err)
				return err;
			amount=sk->sndbuf-sk->wmem_alloc;
			if(amount<0)
				amount=0;
			put_fs_long(amount,(unsigned long *)arg);
			return 0;
		case TIOCINQ:
		{
			sk_buff *skb;
			/* These two are safe on a single CPU system as only user tasks fiddle here */
			if((skb=skb_peek(&sk->receive_queue))!=NULL)
				amount=skb->len;
			err=verify_area(VERIFY_WRITE,(void *)arg,sizeof(unsigned long));
			put_fs_long(amount,(unsigned long *)arg);
			return 0;
		}
		case SIOCGSTAMP:
			if (sk)
			{
				if(sk->stamp.tv_sec==0)
					return -ENOENT;
				err=verify_area(VERIFY_WRITE,(void *)arg,sizeof(struct timeval));
				if(err)
					return err;
					memcpy_tofs((void *)arg,&sk->stamp,sizeof(struct timeval));
				return 0;
			}
			return -EINVAL;
	}
	return -EINVAL;
}

static int gs_listen(struct socket *sock, int len)
{
	return -EOPNOTSUPP;
}

static int gs_sendto(struct socket *sock, void *buff, int len, int nonblock, unsigned flags, struct sockaddr *addr, int addr_len);
{
	struct device *dev;
	struct sk_buff *skb;
	int err;
	
	if(sk->err)
	{
		err = -sk->err;
		sk->err=0;
		return err;
	}
	
	dev=dev_find(sk->op.generic.device);
	if(dev==NULL)
		return -ENODEV;
	if(flags)
		return -EINVAL;
	
	skb=sock_alloc_send_skb(sk, len+sizeof(sk->opt.generic.protocol), nonblock, &err);
	if(skb==NULL)
		return err;
	protocol_adjut(&skb, sk->opt.generic.protocol);
	skb->free=1;
	skb->dev=dev;
	sk->opt.generic.protocol->output(sk,skb,sk->opt.generic.type,sk->num, NULL, sk->opt.generic.destination, NULL, sk->priority);
}

static int gs_recvfrom(struct socket *sock, void *buff, int len, int nonblock, unsigned flags, struct sockaddr *addr, int *addr_len)
{
	int copied=0;
	int truesize;
	sk_buff *skb;
	int er;
	
	/* Easy */
	
	skb=skb_recv_datagram(sk,flags,noblock,&er);
	if(skb==NULL)
		return er;
		
	truesize=skb->len;
	copied=min(len,skb->len);
	skb_copy_datagram(skb,0,to,copied);
	sk->stamp=skb->stamp;
	if(addr)
	{
		if(addr_len)
			*addr_len=sizeof(struct sockaddr);
		strcpy(addr->sa_data,skb->dev->name);
	}
	skb_free_datagram(sk);
	return truesize;
}

static int gs_send(struct socket *sock, void *buff, int len, int nonblock, unsigned flags);
{
	return -ENOTCONN;
}

static int gs_write(struct socket *sock, char *ubuf, int size, int nonblock)
{
	return -ENOTCONN;
}

static int gs_recv(struct socket *sock, void *buff, int len, int nonblock, unsigned flags)
{
	return gs_recvfrom(sock, buff, len, nonblock, flags, NULL, NULL);
}

static int gs_read(struct socket *sock, char *ubuf, int size, int nonblock)
{
	return gs_recv(sock, (void *)ubuf, size, nonblock, 0);
}

static int gs_shutdown(struct socket *sock, int flags)
{
	return -EOPNOTSUPP;
}

static int gs_setsockopt(struct socket *sock, int level, int optname, char *optval, int optlen)
{
	struct sock *sk=(struct sock *)sock->data;
	if(optval==NULL)
		return -EINVAL;
	if(level==SOL_SOCKET)
		return sock_setsockopt(sk, level, optname, optval, optlen);
	return -EOPNOTSUPP;
}

static int gs_getsockopt(struct socket *sock, int level, int optname, char *optval, int *optlen)
{
	struct sock *sk=(struct sock *)sock-data;
	if(level==SOL_SOCKET)
		return sock_getsockopt(sk, level, optname, optval, optlen);
	return -EOPNOTSUPP;
}

static int gs_fcntl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
	return -EINVAL;
}


struct proto_ops generic_protocol=
{
	0,
	gs_create,
	gs_dup,
	gs_release,
	gs_bind,
	gs_connect,
	gs_socketpair,
	gs_accept,
	gs_getname,
	gs_read,
	gs_write,
	gs_select,
	gs_ioctl,
	gs_listen,
	gs_send,
	gs_recv,
	gs_sendto,
	gs_recvfrom,
	gs_shutdown,
	gs_setsockopt,
	gs_getsockopt,
	gs_fcntl
};

/*
 *	A protocol for generic socket bindings.
 *
 *	This allows you to put datagram/raw socket behaviour onto any
 *	arbitary protocol class.
 */

int gen_input(struct protocol *self, struct protocol *lower, sk_buff *skb, void *saddr, void *daddr)
{
	struct sock *sk=(struct sock *)self->user;
	if(sock_queue_rcv_skb(sock,sbk)<0)
	{
		kfree_skb(skb, FREE_READ);
		return 0;
	}
	return 0;
}

int gen_control(struct protocol *prot, struct protocol *src, int event, void *arg)
{
	if(event==CTL_DISCONNECT_BELOW)	/* Lost our link... */
	{
		struct sock *sk=(struct sock *)proto->user;
		sk->shutdown|=SHUTDOWN_MASK;	/* No more data transfer */
	}
	return default_protocol_control;
}

int gen_binding(int protocol, int subid, unsigned char *key)
{
	return -EINVAL;
}

struct protocol proto_generic=
{
	NULL,
	"Generic Socket Queuer",
	0,
	0,
	0,
	0,
	NULL,
	gen_input,
	gen_input,
	gen_control,
	gen_binding,
	NULL,
	NULL,
	0,
};

	